

// std c/c++ includes
#include <iostream>
#include <fstream>
#include <sstream>
#include <memory>
#include <string>
#include <iomanip>
#include <cstdlib>
#include <unistd.h>
#include <getopt.h>

// grpc includes
#include <grpc++/grpc++.h>
#include "turbo.grpc.pb.h"

// yaml includes
#include <yaml-cpp/yaml.h>

// local includes
#include "option.hpp"

#include "afl-fuzz.hpp"


using namespace std;
using namespace grpc;
using namespace Turbo;

int Usage(int argc, char* argv[], OptionVector_t& upper_opts)
{
	cout << getOption(upper_opts, "pgm-name") << ":" << endl;
	cout << "    --help|-h|--usage|-?               Print this screen"                       << endl;
	cout << "    ping                               Verify that the Turbo server is running" << endl;
	cout << "    boost add (-q) (-i) <name>         Create a new boost (quietly).  CLC: If -i, return id of existing" << endl;
	cout << "    boost list                         List all existing boosts"<< endl;
	cout << "clc boost show <id>                    Show details of a boost" << endl;
	cout << "clc boost delete idname>               Delete an existing boost" << endl;
	cout << "    version add (-q) <boost id> <file> Create a new version of a boost (quietly)" << endl;
	cout << "    version list <id>                  List all existing version of a boost"<< endl;
	cout << "clc version show <id>                  Show details of a version of a boost" << endl;
	cout << "    version delete <id>                Delete an existing version of a boost" << endl;
	cout << "    fuzz --fuzz-config <yaml> --app-config <yaml> --ver-id <ver id>" << endl;
	cout << "                                       Start a fuzzing session for version id" << endl;
	cout << "    log get prep <ver id>              Retrieve the prep logs for a version id " << endl;
	cout << "    log get fuzz <ver id>              Retrieve the fuzzing logs for a version id" << endl;
	cout << "clc log delete prep <ver id>           Delete the prep logs for a version id" << endl;
	cout << "clc log delete fuzz <ver id>           Delete the fuzzing logs for a version id" << endl;
	cout << "    seed add <boost id> <seed.yaml>    Add a seed file to a boost" << endl;
	cout << "clc seed list <boost id>               List all seed files for a boost " << endl;
	cout << "    input import-raw-file <input-name> <on-disk-name> <input.yaml>     " << endl;
	cout << "                                       Edit input.yaml to add raw data from a file." << endl;
	cout << "clc input extract-raw-file <in-yaml-name> <on-disk-name> <input.yaml>     " << endl;
	cout << "                                       Extract in-yaml-name from seed.yaml and write to on-disk-name." << endl;
	cout << "clc input blist <boost id>             List all input files for a boost" << endl;
	cout << "clc input vlist <ver id>               List all input files for a version" << endl;
	cout << "clc input download <input id> <fn>     Add a seed file to a boost" << endl;
	return 0;
}

//
// check server alive.
//
int Ping(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	SETUP_CONN(Empty_t,Empty_t, upper_opts);

	// perform the communication
	const auto status = backend->Ping(&context, request, &reply);

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}

	cout << "Server returned OK" << endl;
	return 0;
}

//
// Add a version to a boost
//
int VersionAdd(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	auto verbose_flag = (int)true; // declared int for getopts

	// parse options
	const struct option long_options[] =
		{
			/* These options set a flag. */
			{"no-verbose"   , no_argument      , &verbose_flag, 0  },
			{"verbose"      , no_argument      , &verbose_flag, 1  },
			{0, 0, 0, 0}
		};
	const auto short_opts = string("q");
	while(1)
	{

		/* getopt_long stores the option index here. */
		auto option_index = 0;
		auto c = getopt_long (argc, argv, short_opts.c_str(), long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
			case 0:
			{
				/* If this option set a flag, do nothing else now. */
				if (long_options[option_index].flag != 0)
					break;

				printf ("option %s", long_options[option_index].name);
				if (optarg)
					printf (" with arg %s", optarg);
				printf ("\n");
				abort();
			}

			case 'q':
			{
				verbose_flag=false;
				break;
			}

			default:
			{
				cout << "Unrecognized option to boost add" << endl;
				Usage(argc,argv, upper_opts);
				return 1;
			}
		}
	}

	if(optind + 2 < argc)
	{
		cout << "Error, version add must specify a file name and boost id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(optind + 2 > argc)
	{
		cout << "Junk at end of version add " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	SETUP_CONN(VersionAddRequest_t,VersionAddReply_t, upper_opts);

	request.set_boost_id(atoi(argv[optind+0]));

	ifstream fin(argv[optind+1], ios::binary);
	if(!fin)
	{
		cout << "Cannot open file: " << argv[optind+1] << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	ostringstream ostrm;
	ostrm << fin.rdbuf();
	const auto &send_contents = ostrm.str();

	request.set_version_contents(send_contents);

	const auto status = backend->VersionAdd(&context, request, &reply);

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}

	if(verbose_flag)
	{
		cout << "Server returned OK" << endl;
		cout << "Version ID: " << reply.version_id() << endl;
	}
	else
	{
		cout << reply.version_id() << endl;
	}
	return 0;
}


// 
// list all versions
//
int VersionList(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, version list must specify boost id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>2)
	{
		cout << "Error, junk at end of version list " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	SETUP_CONN(VersionListRequest_t,VersionInfo_t, upper_opts);

	request.set_boost_id(atoi(argv[1]));

        auto reader = move(unique_ptr<ClientReader<VersionInfo_t> >( backend->VersionList(&context, request)));


        cout << "id\t" 
	     << setw(16) << left << "orig bin sha"
	     << setw(16) << left << "zafl bin sha"
	     << setw(25) << left << "submit time"
	     << setw(25) << left << "processing start time"
	     << setw(25) << left << "processing end time"
	     << "error" 
	     << endl;
        while(reader->Read(&reply))
        {
                cout << dec 
		     << reply.ver_id        ()<< "\t" 
		     << setw(16) << left << reply.orig_bin_sha  ()
		     << setw(16) << left << reply.zafl_bin_sha  ()
		     << setw(25) << left << reply.submit_time   ()
		     << setw(25) << left << reply.start_time    ()
		     << setw(25) << left << reply.end_time      ()
		     << boolalpha << reply.err()           
                     << endl;

        }

        const auto status = reader->Finish();

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}
	cout << "Server returned OK" << endl;
	return 0;
}

//
// show details of a version
//
int VersionShow(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	assert(0);
	return 1;
}

//
// delete a version
//
int VersionDelete(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	assert(0);
	return 1;
}



//
// process the boost subcommand
//
int Version(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	const auto subcommands = SubCommandVector_t
		{
			{"add",    VersionAdd    },
			{"list",   VersionList   },
			{"show",   VersionShow   },
			{"delete", VersionDelete },
		};

	return subcommandDispatch("version", subcommands, argc, argv, upper_opts);
}


//
// add a "boost" to to the server
// 
int BoostAdd(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	auto verbose_flag = (int)true; // declared int for getopts

	// parse options
	const struct option long_options[] =
		{
			/* These options set a flag. */
			{"no-verbose"   , no_argument      , &verbose_flag, 0  },
			{"verbose"      , no_argument      , &verbose_flag, 1  },
			{0, 0, 0, 0}
		};
	const auto short_opts = string("q");
	while(1)
	{

		/* getopt_long stores the option index here. */
		auto option_index = 0;
		auto c = getopt_long (argc, argv, short_opts.c_str(), long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
			case 0:
			{
				/* If this option set a flag, do nothing else now. */
				if (long_options[option_index].flag != 0)
					break;

				printf ("option %s", long_options[option_index].name);
				if (optarg)
					printf (" with arg %s", optarg);
				printf ("\n");
				abort();
			}

			case 'q':
			{
				verbose_flag=false;
				break;
			}

			default:
			{
				cout << "Unrecognized option to boost add" << endl;
				Usage(argc,argv, upper_opts);
				return 1;
			}
		}
	}


	if(optind + 1 < argc)
	{
		cout << "Error, boost add must specify a boost name" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(optind + 1 > argc)
	{
		cout << "Error, junk at end of boost add " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	SETUP_CONN(BoostAddRequest_t,BoostAddReply_t, upper_opts);

	request.set_name(argv[optind+0]);

	const auto status = backend->BoostAdd(&context, request, &reply);

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}

	if(verbose_flag)
	{
		cout << "Server returned OK" << endl;
		cout << "Boost ID: " << reply.boost_id() << endl;
	}
	else
	{
		cout << reply.boost_id() << endl;
	}
	return 0;
}


// 
// list all boosts
//
int BoostList(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	if(argc>1)
	{
		cout << "Error, junk at end of boost list " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	SETUP_CONN(Empty_t,BoostInfo_t, upper_opts);

	auto reader = move(unique_ptr<ClientReader<BoostInfo_t> >( backend->BoostList(&context, request)));

	cout << "BoostID\tCreate Time\t\t\tName" << endl;
	while(reader->Read(&reply))
	{
		cout << dec 
		     << (reply.id())<< "\t" 
		     << (reply.create_time())<< "\t" 
		     << (reply.name()) 
		     << endl;
		       
	}

	const auto status = reader->Finish();

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}
	cout << "Server returned OK" << endl;
	return 0;
}

// 
// show a particular boost
//
int BoostShow(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	assert(0);
	return 1;
}

//
// remove a boost from the system
//
int BoostDelete(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	assert(0);
	return 1;
}

//
// process the boost subcommand
//
int Boost(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	const auto subcommands = SubCommandVector_t 
		{
			{"add",    BoostAdd    },
			{"list",   BoostList   },
			{"show",   BoostShow   },
			{"delete", BoostDelete },
		};

	return subcommandDispatch("boost", subcommands, argc, argv, upper_opts);
}

//
// process the boost subcommand
//
int Fuzz(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, fuzz subcommand requires version id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	// default options
	auto verbose_flag         = (int)false; // declared int for getopts
	auto fuzz_config_filename = string();
	auto app_config_filename  = string();
	auto ver_id               = -1;

	// parse options
	const struct option long_options[] =
		{
			/* These options set a flag. */
			{"verbose"      , no_argument      , &verbose_flag, 1  },
			{"no-verbose"   , no_argument      , &verbose_flag, 0  },

			/* These options don’t set a flag.  We distinguish them by their indices. */
			{"fuzz-config"  , required_argument, nullptr, 'f'},
			{"app-config"   , required_argument, nullptr, 'a'},
			{"ver-id"       , required_argument, nullptr, 'v'},
			{0, 0, 0, 0}
		};
	const auto short_opts = string("f:a:h?");
	while(1)
	{

		/* getopt_long stores the option index here. */
		auto option_index = 0;
		auto c = getopt_long (argc, argv, short_opts.c_str(), long_options, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
			case 0:
			{
				/* If this option set a flag, do nothing else now. */
				if (long_options[option_index].flag != 0)
					break;

				printf ("option %s", long_options[option_index].name);
				if (optarg)
					printf (" with arg %s", optarg);
				printf ("\n");
				abort();
			}

			case 'f':
			{
				fuzz_config_filename=string(optarg);
				break;
			}

			case 'a':
			{
				app_config_filename=string(optarg);
				break;
			}
			case 'v':
			{
				ver_id=atoi(optarg);
				break;
			}

			case '?':
			case 'h':
			{
				Usage(argc,argv, upper_opts);
				return 1;
			}

			default:
			{
				cout << "Unrecognized option to fuzz" << endl;
				Usage(argc,argv, upper_opts);
				return 1;
			}
		}
	}


	//
	// sanity check and load fuzzing config
	// 
	if(fuzz_config_filename == "" )
	{
		cout<< "Must specify a fuzzing config file " << endl;
		Usage(argc,argv, upper_opts);
		return 1;

	}
	if(app_config_filename == "" )
	{
		cout<< "Must specify an application config file " << endl;
		Usage(argc,argv, upper_opts);
		return 1;

	}
	if(ver_id < 0)
	{
		cout<< "Invalid or unspecified version id " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	try
	{
		const auto fuzz_config_yaml=YAML::LoadFile(fuzz_config_filename);
		const auto app_config_yaml=YAML::LoadFile(app_config_filename);
		// 
		// log some stuff before we start actually fuzzing
		//
		// cout << "The fuzz config is: " << fuzz_config_yaml << endl;
		// cout << "The app config is: "  << app_config_yaml  << endl;


		// start fuzzer and get exit code
		const auto fuzz_res = FuzzBase_t::factory(fuzz_config_yaml, app_config_yaml, upper_opts, ver_id)->execute();
		return fuzz_res;
	}
	catch(const exception& e)
	{
		cout << "Could not parse yaml file " << endl;
		cout << e.what() << endl;
		return 1;
	}

	return 1;
}

int LogDelete(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	assert(0);
	return 1;
}

int LogGetPrep(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, log get prep requires version id " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>2)
	{
		cout << "Error, junk at end of log get prep command " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	SETUP_CONN(GetZaflLogRequest_t,LogFile_t, upper_opts);

	request.set_ver_id(atoi(argv[1]));

	const auto status = backend->GetZaflLog(&context, request, &reply);

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}
	if(reply.log() == "" )
	{
		cout << "Log is currently empty." << endl;
	}
	else
	{
		cout << "The current prep. log is: " << endl << reply.log() << endl;
	}
	return 1;
}

int LogGetFuzz(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, log get fuzz requires version id " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>2)
	{
		cout << "Error, junk at end of log get fuzz command " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	SETUP_CONN(GetFuzzLogRequest_t,LogFile_t, upper_opts);

	request.set_ver_id(atoi(argv[1]));

	const auto status = backend->GetFuzzLog(&context, request, &reply);

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}
	if(reply.log() == "" )
	{
		cout << "Log is currently empty." << endl;
	}
	else
	{
		cout << "The current prep. log is: " << endl << reply.log() << endl;
	}
	return 1;
}

int LogGet(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	const auto subcommands = SubCommandVector_t 
		{
			{"prep", LogGetPrep },
			{"fuzz", LogGetFuzz },
		};

	return subcommandDispatch("log get", subcommands, argc, argv, upper_opts);
}

//
// process the boost subcommand
//
int Log(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	const auto subcommands = SubCommandVector_t 
		{
			{"get",    LogGet    },
			{"delete", LogDelete },
		};

	return subcommandDispatch("log", subcommands, argc, argv, upper_opts);
}

//
// Import file into seed on disk
//
int SeedImportRawFile(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	if(argc!=4)
	{
		cout << "seed import-raw-file needs <virtual name> <filename toimport> <seed.yaml>" << endl;
		Usage(argc,argv,upper_opts);
		return 1;
	}
	const auto virtual_filename = argv[1];
	const auto to_import_name   = argv[2];
	const auto seed_yaml_name   = argv[3];
	
	auto seed_yaml=YAML::Node(); 
	try
	{
		seed_yaml=YAML::LoadFile(seed_yaml_name);
	}
	catch(const exception& e)
	{
		cout << "Could not parse " << seed_yaml_name << endl;
		cout << e.what() << endl;
		return 1;
	}

	ifstream fin(to_import_name, ios::binary);
	if(!fin)
	{
		cout << "Cannot open file: " << to_import_name << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	ostringstream ostrm;
	ostrm << fin.rdbuf();
	const auto &to_import_contents = ostrm.str();

	// allocate the new dictionary node
	auto new_file_node        = YAML::Node();
	
	// update the fields
	new_file_node["name"]     = virtual_filename;
	auto binary_data_node     = YAML::Binary(reinterpret_cast<const unsigned char*>(to_import_contents.c_str()), to_import_contents.size());
	new_file_node["contents"] = binary_data_node;;
	new_file_node["perms"]    = "rw";

	// add the new dictionary node to the files node.
	auto files_node           = seed_yaml["files"];
	files_node.push_back(new_file_node);

	// write the yaml back out.
	ofstream fout(seed_yaml_name);
	fout << seed_yaml;

	return 0;
}

int SeedAdd(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	if(argc<3)
	{
		cout << "Error, seed add must specify a file name and boost id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>3)
	{
		cout << "Error, junk at end of seed add " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	const auto boost_id      = atoi(argv[1]);
	const auto seed_filename = string(argv[2]);

	ifstream fin(seed_filename, ios::binary);
	if(!fin)
	{
		cout << "Cannot open file: " << argv[2] << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	ostringstream ostrm;
	ostrm << fin.rdbuf();
	const auto &send_contents = ostrm.str();

	SETUP_CONN(SeedAddRequest_t,SeedAddReply_t, upper_opts);
	request.set_boost_id(boost_id);
	request.set_seed_file(send_contents);

	const auto status = backend->SeedAdd(&context, request, &reply);

	// Act upon its status.
	if (!status.ok()) 
	{
		print_error(status);
		return 1;
	}

	cout << "Server returned OK" << endl;
	cout << "Input ID: " << reply.input_id() << endl;
	cout << "Seed ID : " << reply.seed_id()  << endl;
	return 0;
}


//
// process the boost subcommand
//
int Seed(int argc, char *argv[], OptionVector_t& upper_opts) 
{
	const auto subcommands = SubCommandVector_t 
		{
			{"add",             SeedAdd        },
			{"import-raw-file", SeedImportRawFile },
		};

	return subcommandDispatch("seed", subcommands, argc, argv, upper_opts);
}

// 
// handle processing of the turbo cli
//
int main(int argc, char *argv[])
{
	const auto subcommands = SubCommandVector_t 
		{
			{"ping",    Ping    },
			{"boost",   Boost   },
			{"version", Version },
			{"log",     Log     },
			{"fuzz",    Fuzz    },
			{"seed",    Seed    },
			{"--help",  Usage   },
			{"help",    Usage   },
			{"--usage", Usage   },
			{"usage",   Usage   },
			{"-h",      Usage   },
			{"-?",      Usage   },
		};
	auto options_vector = OptionVector_t
		{
			{"pgm-name", argv[0]}, 
			{"host", "localhost:55155"}
		};
	return subcommandDispatch("turbo-cli", subcommands, argc, argv, options_vector);
}
