
#include "afl-fuzz.hpp"
#include "parse-utils.hpp"


// include experimental filesystem support
#include <experimental/filesystem>
namespace fs = std::experimental::filesystem;


using namespace Turbo;

        //
        // how to download a zafld file.
        //
const int32_t download_file(const string& filename, const uint32_t ver_id, const OptionVector_t& opts)
{

	auto zafld_bin = string();
	while(true)
	{
		SETUP_CONN(GetZafldRequest_t,ZafldBinary_t, opts);

		request.set_ver_id(ver_id);
		const auto status = backend->GetZafld(&context, request, &reply);

		// Act upon its status.
		if (!status.ok())
		{
			print_error(status);
			return 1;
		}
		if(reply.err())
		{
			cout << "Detected error in creating fuzz-ready version" << endl;
			return 1;
		}
		if(reply.zafld_bin() != "" )
		{
			zafld_bin = reply.zafld_bin();
			break;
		}
		cout << "Binary not ready, sleeping 5 seconds..." << endl;
		sleep(5);
	}
	cout << "Server returned OK on getting Zafl'd binary with size " << zafld_bin.size() << endl;

	ofstream strm(filename, ios::binary);
	strm << zafld_bin;
	strm.close();
	fs::permissions(filename, fs::perms::owner_all | fs::perms::add_perms);
	return 0;
};


int32_t AflFuzz_t::getAllInputs()
{
	const auto base_res = FuzzBase_t::getAllInputs(); 	
	if(base_res!=0)
		return base_res;

	for(auto input : inputs)
	{
		const auto &yaml = input.yaml;
		if(! yaml["files"])
		{
			cout << "Yaml has no files? " << endl;
			return 1;
		}
		if(! yaml["files"].IsSequence() ) 
		{
			cout <<" Yaml's files section isn't a sequence? " << endl;
			return 1;
		}

		auto input_files = FlatInput_t();
		for(auto file : yaml["files"])
		{
			if(!file.IsMap())
			{
				cout << "Found file which is not a map? " << endl;
				return 1;
			}

			const auto& file_contents = file["contents"].as<string>() ;

			cout << "Found file named: " << file["name"].as<string>() << " of size " << file_contents.size() << endl;
			
			input_files.push_back(file_contents);
		}
		flat_input_to_id_map[input_files]=input.id;
		flat_inputs.insert(input_files);
	}

	// for now, return err
	return 1;
}

void AflFuzz_t::cleanWorkdir()
{
	const auto &dir = fuzz_config["working-dir"].as<string>();

	// supported by c++17, but not by gcc7 with --std=c++17.   too bad.
	(void)fs::remove_all(dir);
	(void)fs::create_directories(dir+"/in");
	(void)fs::create_directories(dir+"/out");
}

void AflFuzz_t::writeInputs()
{
	const auto &dir = fuzz_config["working-dir"].as<string>();
	auto i = 0u;
	for(const auto& flat_input : flat_inputs)
	{
		assert(flat_input.size() == 1);
		for(const auto& flat_file : flat_input)
		{
			ofstream fout(dir+"/in/"+to_string(i++));
			fout << flat_file;
		}
	}
}


set<string> AflFuzz_t::generateCmdLines() const
{
	auto cmd_lines = set<string> ();
	for(auto input : inputs)
	{
		const auto &yaml = input.yaml;
		auto cmd_line = string();
		for(const auto& arg_node : yaml["cmdline"])
		{
			if(arg_node.IsScalar())
			{
				cmd_line += " " + arg_node.as<string>();
			}
			else if(arg_node.IsMap())
			{
				const auto fn_node = arg_node["filename"];
				if(fn_node.IsScalar())
				{
					const auto fn_name = fn_node.as<string>();
					if(fn_name != "stdin")
						cmd_line += " @@";


				}
				else
				{
					stringstream strm;
					strm << "Unknown node type in cmd_line: " << fn_node;
					throw invalid_argument(strm.str());
				}
			}
			else
			{
				stringstream strm;
				strm << "Unknown node type in cmd_line: " << arg_node;
				throw invalid_argument(strm.str());
			}
		}
		cmd_lines.insert(cmd_line);

	}
	return cmd_lines;
}

void AflFuzz_t::handleAflGeneratedInputs()
{
	const auto workingDir = fuzz_config["working-dir"].as<string>();
	const auto path = workingDir + "/out/queue";
	for (const auto& entry : fs::directory_iterator(path))
	{
		(void)uploadInput(entry.path());
	}
}

int32_t AflFuzz_t::uploadInput(const string& path)
{
	ifstream fin(path);
	stringstream strm;
	strm << fin.rdbuf();
	const auto& file_contents = strm.str();

	cout << " Size: " << file_contents.size() << endl;

	const auto flat_input = FlatInput_t{file_contents};
	const auto flat_input_it = flat_inputs.find(flat_input);
	if(flat_input_it == flat_inputs.end())
	{
		// need to convert to yaml and upload
		auto new_yaml = YAML::Clone(inputs[0].yaml);
		new_yaml["files"][0]["contents"]=YAML::Binary(reinterpret_cast<const unsigned char*>(file_contents.c_str()), file_contents.size());
		stringstream strm; 
		strm << new_yaml;

		SETUP_CONN(InputAddRequest_t, InputAddReply_t, options);
		request.set_ver_id(ver_id);
		request.set_input_file(strm.str());
		const auto status = backend->InputAdd(&context,request,&reply);
		if (!status.ok())
		{
			print_error(status);
			throw invalid_argument("Could not upload input");
		}
		cout << "Uploaded as input_id = " << reply.input_id() << endl;
		return reply.input_id();
	}
	return flat_input_to_id_map[flat_input];

}

void AflFuzz_t::uploadCrashes(const string& sub_dir)
{
	const auto workingDir = fuzz_config["working-dir"].as<string>();
	const auto path = workingDir + "/" + sub_dir;
	for (const auto& entry : fs::directory_iterator(path))
	{
		const auto input_id = uploadInput(entry.path());
		SETUP_CONN(AddForensicsRequest_t, AddForensicsReply_t, options);
		request.set_ver_id  (ver_id);
		request.set_input_id(input_id);
		const auto status = backend->AddForensics(&context,request,&reply);
		if (!status.ok())
		{
			print_error(status);
			throw invalid_argument("Could not upload input");
		}
		cout << "Uploaded as forensics as id = " << reply.forensic_id() << endl;
	}
}


string AflFuzz_t::getAflCmd()
{
	cout << "Fuzz time is: " ;
	const auto fuzzing_time = to_seconds_duration(fuzz_config["fuzzing-time"]);
	cout << fuzzing_time << " seconds" << endl;

	cout << "Input timeout is: " ;
	const auto input_timeout = to_seconds_duration(fuzz_config["input-timeout"]);
	cout << fuzzing_time << " seconds" << endl;

	const auto working_dir = fuzz_config["working-dir"].as<string>();

	auto cmd = string() + "cd " + working_dir + " ; ";
	cmd += "timeout " + to_string(fuzzing_time) + " afl-fuzz -i in -o out ";
	cmd += "-t " + to_string(input_timeout*1000);
	cmd += "-- /tmp/to_fuzz.exe ";


	const auto cmd_lines = generateCmdLines();
	if(cmd_lines.size() != 1)
	{
		// todo:  check for back-end specific flag to say which cmd-line to fuzz.
		throw invalid_argument( "AFL backend must use exactly 1 cmd line. ");
	}

	// append first cmd line
	cmd+=*(cmd_lines.begin());

	return cmd;

}

void AflFuzz_t::doAfl()
{
	const auto afl_cmd = getAflCmd();

	cout << "Afl cmd is: " << afl_cmd << endl;

        const auto pipe = popen(afl_cmd.c_str(), "r");
        if (!pipe)
        {
                throw invalid_argument(string()+"Couldn't start command: " + afl_cmd); 
        }

	cout << "The AFL log is: " << endl;


	auto log_buffer = string();
        auto last_update = time(nullptr);
        auto c=EOF;
        do
        {
                // get the next byte from the pipe. this may block.
                auto fread_res=fread(&c,1,1,pipe);
                if(fread_res != 1) break;
		cout << (char)c;
		log_buffer += (char)c;

                const auto log_buff_big          = log_buffer.size() > 1024;
                const auto too_long              = difftime(time(nullptr), last_update) > 5;
                const auto non_zero_buf          = log_buffer.size() > 0;
                const auto too_long_and_non_zero = too_long && non_zero_buf;
                if(log_buff_big || too_long_and_non_zero)
                {
			// upload our log
			SETUP_CONN(AddFuzzLogRequest_t,Empty_t,options);
			request.set_ver_id(ver_id);
			request.set_log(log_buffer);
			const auto status = backend->AddFuzzLog(&context,request,&reply);
			if (!status.ok())
			{
				print_error(status);
			}
			log_buffer="";
			last_update=time(nullptr);
                }
	} while ( c != EOF );

	const auto returnCode = pclose(pipe);
        cout << "Afl returned: " << returnCode << endl;
}

int32_t AflFuzz_t::execute()
{
        //
        // download the zafld file and check for err.
        //
        const auto df_res = download_file("/tmp/to_fuzz.exe", ver_id, options);
	if(df_res!=0)
		return 1;


	cout << "Working dir is: ";
	const auto working_dir = fuzz_config["working-dir"].as<string>();
	cout << working_dir << endl;

	getAllInputs();
	cleanWorkdir();
	writeInputs();
	doAfl();
	handleAflGeneratedInputs();
	uploadCrashes("/out/crashes");
	uploadCrashes("/out/hangs");

	return 0; // err
}
